/*	XGPostError.cpp
 *
 *		This is the code that handles posting an error alert box.
 *	This contains the appropriate code for popping up an alert
 *	box in Windows, on the Macintosh, and with X Windows
 */

/**********************************************************************

  YAAF is Copyright 1997-1999 by William Woody.

  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met:

  Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

  Neither the name of the copyright holder or other contributors may
  be used to promote products derived from this software without
  specific written prior permission.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS 
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN 
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
  SOFTWARE.
  
  Contact:   woody@alumni.caltech.edu
  Web page:  http://www.pandawave.com/yaaf/

**********************************************************************/

#include <XConfig.h>			// must be first to optimize compile on MSVC++
#include <XError.h>
#include <XDataUtil.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <XDialog.h>

/************************************************************************/
/*																		*/
/*	Globals																*/
/*																		*/
/************************************************************************/

char XGPostError::gSMessage[256];
char XGPostError::gLMessage[256];
bool XGPostError::gResult;

/************************************************************************/
/*																		*/
/*	Standard Construction												*/
/*																		*/
/************************************************************************/

/*$	XGPostError::XGPostError
 *
 *		Three versions. The first is used by children classes which
 *	inherit from me
 */

XGPostError::XGPostError()
{
	fType = KXGAlertFatal;
}

/*$	XGPostError::XGPostError
 *
 *		Second version. This version is used when the error is created
 *	by most applications. To use this, to throw an error, you would
 *	write 'throw XGPostError(KErrorID,args,...);'
 */

XGPostError::XGPostError(unsigned long msg, ...)
{
	va_list vlist;
	char bufs[256];
	char bufl[256];
	
	::GetResourceMessage(GETHIWORD(msg),GETLOWORD(msg),bufs,bufl);
	
	fType = KXGAlertFatal;
	va_start(vlist,msg);
	vyformat(gSMessage,bufs,vlist);
	vyformat(gLMessage,bufl,vlist);
	va_end(vlist);
}

/*$	XGPostError::XGPostError
 *
 *		Third version; this is the varargs version of the second
 *	constructor, and is used internally by YAAF.
 */

XGPostError::XGPostError(unsigned long msg, va_list vlist)
{
	fType = KXGAlertFatal;
	char bufs[256];
	char bufl[256];
	
	::GetResourceMessage(GETHIWORD(msg),GETLOWORD(msg),bufs,bufl);
	vyformat(gSMessage,bufs,vlist);
	vyformat(gLMessage,bufl,vlist);
}

/************************************************************************/
/*																		*/
/*	Construction Support												*/
/*																		*/
/************************************************************************/

/*$	XGPostError::SetMessage
 *
 *		Set the message. If no message is specified by the
 *	constructor, this is used to initialize the message. This
 *	is used by the constructor of classes which descend from the
 *	XGPostError class.
 */

void XGPostError::SetMessage(unsigned long msg,...)
{
	va_list vlist;
	char bufs[256];
	char bufl[256];
	
	va_start(vlist,msg);
	::GetResourceMessage(GETHIWORD(msg),GETLOWORD(msg),bufs,bufl);
	vyformat(gSMessage,bufs,vlist);
	vyformat(gLMessage,bufl,vlist);
	va_end(vlist);
}

/*$	XGPostError::SetMessage
 *
 *		Set the message. This is the varargs variation of the above.
 */

void XGPostError::SetMessage(unsigned long msg, va_list vlist)
{
	char bufs[256];
	char bufl[256];

	::GetResourceMessage(GETHIWORD(msg),GETLOWORD(msg),bufs,bufl);
	vyformat(gSMessage,bufs,vlist);
	vyformat(gLMessage,bufl,vlist);
}

/*$	XGPostError::GetShortError
 *
 *		Get the actual error message. This returns the short error
 *	message.
 */

void XGPostError::GetShortError(char *dest)
{
	strcpy(dest,gSMessage);
}

/*$	XGPostError::GetLongError
 *
 *		Get the actual error message. This returns the long description
 *	of the error.
 */

void XGPostError::GetLongError(char *dest)
{
	strcpy(dest,gLMessage);
}

/************************************************************************/
/*																		*/
/*	Error Processing													*/
/*																		*/
/************************************************************************/

/*	Dialog resources
 *
 *		These are the resources that are used for constructing my
 *	different dialog alerts
 */

static unsigned char GDialogError[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5C, /* 00000000: .......\ */
    0x00, 0x00, 0x00, 0x4C, 0x64, 0x6C, 0x6F, 0x67, /* 00000008: ...Ldlog */
    0x00, 0x00, 0x00, 0x12, 0x00, 0x00, 0x00, 0x00, /* 00000010: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000018: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000020: ........ */
    0x00, 0x00, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000028: .....J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000030: .v...J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000038: .v...J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000040: .v...J.. */
    0x00, 0x76, 0x05, 0x45, 0x72, 0x72, 0x6F, 0x72, /* 00000048: .v.Error */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x64, /* 00000050: .......d */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x42, /* 00000058: .......B */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x32, /* 00000060: .......2 */
    0x69, 0x63, 0x6F, 0x6E, 0x00, 0x00, 0x00, 0x00, /* 00000068: icon.... */
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000070: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, /* 00000078: ........ */
    0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x38, /* 00000080: .......8 */
    0x00, 0x00, 0x00, 0x30, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000088: ...0.... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000090: ........ */
    0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, /* 00000098: ........ */
    0x00, 0x3E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000A0: .>...... */
    0x00, 0x2E, 0x74, 0x65, 0x78, 0x74, 0x00, 0x00, /* 000000A8: ..text.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 000000B0: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 000000B8: ........ */
    0x00, 0x50, 0x00, 0x00, 0x00, 0x0E, 0xFF, 0xFF, /* 000000C0: .P...... */
    0xFF, 0xF2, 0x00, 0x00, 0x00, 0x1E, 0xFF, 0xFF, /* 000000C8: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x02, 0x5E, 0x30, 0x00, /* 000000D0: .....^0. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3E, /* 000000D8: .......> */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2E, /* 000000E0: ........ */
    0x74, 0x65, 0x78, 0x74, 0x00, 0x00, 0x00, 0x00, /* 000000E8: text.... */
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, /* 000000F0: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5A, /* 000000F8: .......Z */
    0x00, 0x00, 0x00, 0x1E, 0xFF, 0xFF, 0xFF, 0xF2, /* 00000100: ........ */
    0x00, 0x00, 0x00, 0x50, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000108: ...P.... */
    0x00, 0x00, 0x02, 0x5E, 0x31, 0x00, 0x00, 0x00, /* 00000110: ...^1... */
    0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000118: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x32, 0x70, 0x62, /* 00000120: .....2pb */
    0x74, 0x6E, 0x00, 0x00, 0x00, 0x64, 0x00, 0x00, /* 00000128: tn...d.. */
    0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 00000130: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0x00, 0x00, /* 00000138: ........ */
    0x00, 0x56, 0x00, 0x00, 0x01, 0x42, 0x00, 0x00, /* 00000140: .V...B.. */
    0x00, 0x72, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000148: .r...... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000150: ........ */
    0x02, 0x4F, 0x4B, 0x00                          /* 00000158: .OK.     */
}; /* _R56526573_0082 */

static unsigned char GDialogCaution[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5E, /* 00000000: .......^ */
    0x00, 0x00, 0x00, 0x4E, 0x64, 0x6C, 0x6F, 0x67, /* 00000008: ...Ndlog */
    0x00, 0x00, 0x00, 0x12, 0x00, 0x00, 0x00, 0x00, /* 00000010: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000018: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000020: ........ */
    0x00, 0x00, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000028: .....J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000030: .v...J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000038: .v...J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000040: .v...J.. */
    0x00, 0x76, 0x07, 0x43, 0x61, 0x75, 0x74, 0x69, /* 00000048: .v.Cauti */
    0x6F, 0x6E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000050: on...... */
    0x00, 0x64, 0x00, 0x00, 0x00, 0x65, 0x00, 0x00, /* 00000058: .d...e.. */
    0x00, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000060: .B...... */
    0x00, 0x32, 0x69, 0x63, 0x6F, 0x6E, 0x00, 0x00, /* 00000068: .2icon.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 00000070: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000078: ........ */
    0x00, 0x18, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, /* 00000080: ........ */
    0x00, 0x38, 0x00, 0x00, 0x00, 0x30, 0xFF, 0xFF, /* 00000088: .8...0.. */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, /* 00000090: ........ */
    0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, /* 00000098: ........ */
    0x00, 0x00, 0x00, 0x3E, 0x00, 0x00, 0x00, 0x00, /* 000000A0: ...>.... */
    0x00, 0x00, 0x00, 0x2E, 0x74, 0x65, 0x78, 0x74, /* 000000A8: ....text */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000B0: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 000000B8: ........ */
    0x00, 0x00, 0x00, 0x50, 0x00, 0x00, 0x00, 0x0E, /* 000000C0: ...P.... */
    0xFF, 0xFF, 0xFF, 0xF2, 0x00, 0x00, 0x00, 0x1E, /* 000000C8: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x02, 0x5E, /* 000000D0: .......^ */
    0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000D8: 0....... */
    0x00, 0x3E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000E0: .>...... */
    0x00, 0x2E, 0x74, 0x65, 0x78, 0x74, 0x00, 0x00, /* 000000E8: ..text.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 000000F0: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 000000F8: ........ */
    0x00, 0x5A, 0x00, 0x00, 0x00, 0x1E, 0xFF, 0xFF, /* 00000100: .Z...... */
    0xFF, 0xF2, 0x00, 0x00, 0x00, 0x50, 0xFF, 0xFF, /* 00000108: .....P.. */
    0xFF, 0xFF, 0x00, 0x00, 0x02, 0x5E, 0x31, 0x00, /* 00000110: .....^1. */
    0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x42, /* 00000118: .......B */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x32, /* 00000120: .......2 */
    0x70, 0x62, 0x74, 0x6E, 0x00, 0x00, 0x00, 0x64, /* 00000128: pbtn...d */
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000130: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFA, /* 00000138: ........ */
    0x00, 0x00, 0x00, 0x56, 0x00, 0x00, 0x01, 0x3E, /* 00000140: ...V...> */
    0x00, 0x00, 0x00, 0x72, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000148: ...r.... */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000150: ........ */
    0x00, 0x00, 0x02, 0x4F, 0x4B, 0x00, 0x00, 0x00, /* 00000158: ...OK... */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000160: ........ */
    0x00, 0x36, 0x70, 0x62, 0x74, 0x6E, 0x00, 0x00, /* 00000168: .6pbtn.. */
    0x00, 0x65, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 00000170: .e...... */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000178: ........ */
    0x00, 0xAE, 0x00, 0x00, 0x00, 0x56, 0x00, 0x00, /* 00000180: .....V.. */
    0x00, 0xF2, 0x00, 0x00, 0x00, 0x72, 0xFF, 0xFF, /* 00000188: .....r.. */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 00000190: ........ */
    0x00, 0x00, 0x00, 0x00, 0x06, 0x43, 0x61, 0x6E, /* 00000198: .....Can */
    0x63, 0x65, 0x6C, 0x00                          /* 000001A0: cel.     */
}; /* _R56526573_0081 */

static unsigned char GDialogWarning[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x5E, /* 00000000: .......^ */
    0x00, 0x00, 0x00, 0x4E, 0x64, 0x6C, 0x6F, 0x67, /* 00000008: ...Ndlog */
    0x00, 0x00, 0x00, 0x12, 0x00, 0x00, 0x00, 0x00, /* 00000010: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000018: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000020: ........ */
    0x00, 0x00, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000028: .....J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000030: .v...J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000038: .v...J.. */
    0x00, 0x76, 0x00, 0x00, 0x01, 0x4A, 0x00, 0x00, /* 00000040: .v...J.. */
    0x00, 0x76, 0x07, 0x57, 0x61, 0x72, 0x6E, 0x69, /* 00000048: .v.Warni */
    0x6E, 0x67, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000050: ng...... */
    0x00, 0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000058: .d...... */
    0x00, 0x42, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000060: .B...... */
    0x00, 0x32, 0x69, 0x63, 0x6F, 0x6E, 0x00, 0x00, /* 00000068: .2icon.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 00000070: ........ */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000078: ........ */
    0x00, 0x18, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, /* 00000080: ........ */
    0x00, 0x38, 0x00, 0x00, 0x00, 0x30, 0xFF, 0xFF, /* 00000088: .8...0.. */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, /* 00000090: ........ */
    0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, /* 00000098: ........ */
    0x00, 0x00, 0x00, 0x3E, 0x00, 0x00, 0x00, 0x00, /* 000000A0: ...>.... */
    0x00, 0x00, 0x00, 0x2E, 0x74, 0x65, 0x78, 0x74, /* 000000A8: ....text */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000B0: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, /* 000000B8: ........ */
    0x00, 0x00, 0x00, 0x50, 0x00, 0x00, 0x00, 0x0E, /* 000000C0: ...P.... */
    0xFF, 0xFF, 0xFF, 0xF2, 0x00, 0x00, 0x00, 0x1E, /* 000000C8: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x02, 0x5E, /* 000000D0: .......^ */
    0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000D8: 0....... */
    0x00, 0x3E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 000000E0: .>...... */
    0x00, 0x2E, 0x74, 0x65, 0x78, 0x74, 0x00, 0x00, /* 000000E8: ..text.. */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, /* 000000F0: ........ */
    0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, /* 000000F8: ........ */
    0x00, 0x5A, 0x00, 0x00, 0x00, 0x1E, 0xFF, 0xFF, /* 00000100: .Z...... */
    0xFF, 0xF2, 0x00, 0x00, 0x00, 0x50, 0xFF, 0xFF, /* 00000108: .....P.. */
    0xFF, 0xFF, 0x00, 0x00, 0x02, 0x5E, 0x31, 0x00, /* 00000110: .....^1. */
    0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, /* 00000118: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x32, /* 00000120: .......2 */
    0x70, 0x62, 0x74, 0x6E, 0x00, 0x00, 0x00, 0x64, /* 00000128: pbtn...d */
    0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000130: ........ */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xF8, /* 00000138: ........ */
    0x00, 0x00, 0x00, 0x54, 0x00, 0x00, 0x01, 0x3A, /* 00000140: ...T...: */
    0x00, 0x00, 0x00, 0x70, 0xFF, 0xFF, 0xFF, 0xFF, /* 00000148: ...p.... */
    0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00000150: ........ */
    0x00, 0x00, 0x02, 0x4F, 0x4B, 0x00              /* 00000158: ...OK.   */
}; /* _R56526573_0080 */


/*$	XGPostError::DisplayError
 *
 *		Do the alert message using the YAAF dialog code.
 */

void XGPostError::DisplayError(void)
{
	long result;
	char bufs[256];
	char bufl[256];
	unsigned char *data;
	
	switch (fType) {
		case KXGAlertFatal:
			data = GDialogError;
			break;
		case KXGAlertCaution:
			data = GDialogCaution;
			break;
		case KXGAlertWarning:
			data = GDialogWarning;
			break;
		default:
			return;			// illegal value
	}
	
	GetShortError(bufs);
	GetLongError(bufl);
	gResult = false;
	result = XGDialog::ModalAlert(data,bufs,bufl,NULL,NULL);
	if ((result == 101) && (fType == KXGAlertCaution)) {
		gResult = true;
	}
}

